/* Parameters.cxx, parses parameters and inputs from constraint files
 *
 * Author: Feng Qian(fqian@sable.mcgill.ca)
 *         Marc Berndl(berndl@sable.mcgill.ca)
 *         Ondrej Lhotak(olhotak@sable.mcgill.ca)
 */

#include "Parameters.h"
#include "helpers.h"

const char Parameters::usageString[] = 
"[USAGE] : ABDD inputFile [n #ofInitialOfNodes]\n\
                          [c cacheSize]\n\
		          [(t)ypes (y)es (n)o]\n\
		          [O ordering]\n\
		          [e inputformat_nul(old|new)]\n\
		          [l controlcode(ite|part)]\n\
		          [a aliasstrategy(merge|sep|none|flow)\n\
		          [i incremental optimization (y)es (n)o]\n\
		          [r reversed local ordering (y)es (n)o]\n";

bool Parameters::thisShouldBePrinted(VerbageLevel importance){
  if(detailLevel == silent) 
    return false;
  
  if(detailLevel==debug) 
    return true;
  
  if(detailLevel==normal && importance==normal)
    return true;
  
  if(detailLevel==verbose){
    if(importance==normal || importance == verbose){
      return true;
    }
  }
  return false;
}

void Parameters::output(char * key, char * string, VerbageLevel importance){
  if(thisShouldBePrinted(importance)){
    (*outputFile) << key << ": " << string << "\n";
  }
}

void Parameters::output(char * key, int value, VerbageLevel importance){
  if(thisShouldBePrinted(importance)){
    (*outputFile) << key << ": " << value << "\n";
  }
}

ostream *Parameters::outputFile= &cout;
VerbageLevel Parameters::detailLevel=defaultDetailLevel;

VarOrdering determineOrdering(char *order) {
  if (!strcmp("ordered", order)) {
    return ORDERED;
  } else if (!strcmp("reversed", order)) {
    return REVERSED;
  } else if (!strcmp("fd_v1_v2_v3_hp", order)) {
    return FD_V1_V2_V3_HP;
  } else if (!strcmp("fd_v1v2v3_hp", order)) {
    return FD_V1xV2xV3_HP;
  } else if (!strcmp("fd_v1v2_v3_hp", order)) {
    return FD_V1xV2_V3_HP;
  } else if (!strcmp("v1v2v3_fd_hp", order)) {
    return V1xV2xV3_FD_HP;
  } else if (!strcmp("hp_v1v2v3_fd", order)) {
    return HP_V1xV2xV3_FD;
  } else if (!strcmp("fdv1v2v3hp", order)) {
    return FDxV1xV2xV3xHP;
  } else if (!strcmp("hp_v1v2v3fd", order)) {
    return HP_V1xV2xV3xFD;
  } else if (!strcmp("v1v2v3fd_hp", order)) { 
    return V1xV2xV3xFD_HP;
  } else if (!strcmp("v1v2fd_hp_v3", order)) { 
    return V1xV2xFD_HP_V3;
  } else if (!strcmp("v1_v2fdhpv3", order)) {
    return V1_V2xFDxHPxV3;
  } else if (!strcmp("v1v2v3_fdhp", order)) {
    return V1xV2xV3_FDxHP;
  } else if (!strcmp("fd_v1v2v3hp", order)) {
    return FD_V1xV2xV3xHP;
  } else if (!strcmp("random", order)) {
    return RANDOM;
  } else if (!strcmp("randvars", order)) {
    return RANDVARS;
  } else if (!strcmp("randfd", order)) {
    return RANDFD;
  } else if (!strcmp("randhp", order)) {
    return RANDHP;
  } else if (!strcmp("randall", order)) {
    return RANDALL;

	//for Solver-noalias
  } else if (!strcmp("fd_v1_v2_h1_h2", order)) {
    return FD_V1_V2_H1_H2;
  } else if (!strcmp("fd_v1v2_h1_h2", order)) {
    return FD_V1xV2_H1_H2;
  } else if (!strcmp("fd_v2_v1_h1_h2", order)) {
    return FD_V2_V1_H1_H2;
  } else if (!strcmp("fd_v2v1_h1_h2", order)) {
    return FD_V2xV1_H1_H2;
  } else if (!strcmp("fdv1v2h1h2", order)) {
    return FDxV1xV2xH1xH2;
  } else if (!strcmp("fd_h1_v1v2_h2", order)) {
    return FD_H1_V1xV2_H2;
  } else if (!strcmp("fd_h1_v2v1_h2", order)) {
    return FD_H1_V2xV1_H2;
  } else if (!strcmp("h1_v1v2_fd_h2", order)) {
    return H1_V1xV2_FD_H2;
  } else if (!strcmp("h1_v2v1_fd_h2", order)) {
    return H1_V2xV1_FD_H2;
  } else if (!strcmp("fd_v1v2h1_h2", order)) {
    return FD_V1xV2xH1_H2;
	// end
  } else {
    return DEFAULT;
  }
}

void Parameters::dumpParameters() {
  cout << "Parameters: \n";
  cout <<   "variable_ordering         " << orderingName << "\n";
  if (typeFilter) {
    cout << "using_types               " << "yes" << "\n";
  } else {
    cout << "using_types               " << "no" << "\n";
  }

  switch (ctrlcode) {
  case ITERATION: 
    cout << "control_code              " << "iteration" << "\n";
    break;
  case PARTITIONED:
    cout << "control_code              " << "partitioned" << "\n";
    break;
  default:
    cout << "control_code              " << "unknown" << "\n";
    break;
  }

  switch (aliasstrategy) {
  case MERGING:
    cout << "alias_strategy            " << "merging\n";
    break;
  case SEPARATE:
    cout << "alias_strategy            " << "separate\n";
    break;
  case NOALIAS:
    cout << "alias_strategy            " << "noalias\n";
    break;
  case FLOWONLY:
    cout << "alias_strategy            " << "flowonly\n";
    break;
  }
  switch(incremental){
  case INCREMENTAL:
    cout << "incremental optimizations " << "on\n";
    break;
  case NONINCREMENTAL:
    cout << "incremental optimizations " << "off\n";
    break;
  }
}

Parameters::Parameters(int count, char ** strings){
  outputFileName="";
  numNodes=defaultNumNodes;
  cacheSize=defaultCacheSize;
  typeFilter=defaultTypeFilter;
  numberOfProgramArguments=count;
  programArguments=strings;
  varencoding=defaultVarEncoding;
  orderingName="fd_v1_v2_h1_h2";
  varordering= FD_V1_V2_H1_H2;
  ctrlcode=defaultControlCode;
  aliasstrategy=defaultAliasStrategy;
  incremental = defaultIncrementalization;
  reversedLocalOrdering = defaultReversedLocalOrdering;

  if(count < 2 || ((count % 2)==1)){
    cerr << usageString;
    exit(-1);
  }
  inputFileName = strings[1];	
  for(int i=2;i<numberOfProgramArguments;i+=2){
    Debug("Reading Parameter",programArguments[i]);
    Debug("Reading Parameter",programArguments[i+1]);
    switch(*programArguments[i]){
      // using type
      case 't':
	switch(*programArguments[i+1]){
  	  case 'n':
	    typeFilter=false;
	    break;
	  case 'y':
	    typeFilter=true;
	  default:
	    break;
	}
	break;      
    case 'o':
      // output
      outputFileName=programArguments[i+1];
      outputFile= new fstream();
      ((fstream*)outputFile)->open(outputFileName,ios::out);
      break;
      
    case 'v':
      // verbose level
      switch(*programArguments[i+1]){
      case 's':
	detailLevel=silent;
	break;
      case 'n':
	detailLevel=normal;
	break;
      case 'v':
	detailLevel=verbose;
	break;
      case 'd':
	detailLevel=debug;
	break;
      default:
	cerr << usageString;
	exit(-1);
      }
      break;
    case 'n':
      // initial node size
      numNodes=atoi(programArguments[i+1]);
      break;
    case 'c':
      // initial cache size
      cacheSize=atoi(programArguments[i+1]);
      break;
    case 'O':  
      // initial variable ordering
      orderingName = programArguments[i+1];
      varordering=determineOrdering(programArguments[i+1]);
      break;
    case 's':  
      // solution file
      solutionFileName = programArguments[i+1];
      break;
    case 'e':
      // variable encoding: 0 - old format; 1 - new format cls/meth/var
      if (!strcmp(programArguments[i+1], "old")) { 
	varencoding = ENUMERATION;
      } else if (!strcmp(programArguments[i+1], "new")) {
	varencoding = CLSMETHVAR;
      } else {
	cerr << "  e  old|new  (input encoding)\n";
      }
      break;
    case 'l':
      // control code : 0 - iteration, 3 - partition
      if (!strcmp(programArguments[i+1], "ite")) {
	ctrlcode = ITERATION;
      } else if (!strcmp(programArguments[i+1], "part")) {
	ctrlcode = PARTITIONED;
      } else {
	cerr << " l  ite|part  (control code)\n";
      }
      break;
    case 'i':
      // control code : 0 - iteration, 3 - partition
      if (!strcmp(programArguments[i+1], "n")) {
	incremental = NONINCREMENTAL;
      } else if (!strcmp(programArguments[i+1], "y")) {
	incremental = INCREMENTAL;
      } else {
	cerr << " i  y|n  (incremental)\n";
      }
      break;
    case 'a':
      // strategy for processing aliased edges
      // 0 - merge to edgeset, 1 - propagate separately
      if (!strcmp(programArguments[i+1], "merge")) {
	aliasstrategy = MERGING;
      } else if (!strcmp(programArguments[i+1], "sep")) {
	aliasstrategy = SEPARATE;
      } else if (!strcmp(programArguments[i+1], "none")){
	aliasstrategy = NOALIAS;
      } else if (!strcmp(programArguments[i+1], "flow")){
	aliasstrategy = FLOWONLY;
      } else {
	cerr << " a  merge|sep|none|flow  (alias strategy)\n";
      }
      break;
	case 'r':
	  if (!strcmp(programArguments[i+1], "n")) {
		reversedLocalOrdering = false;
	  }
	  break;
    default:
      cerr << usageString;
      exit(-1);
    }
  }

  dumpParameters();
  
  EXIT("Parameters Init");
}	


FileType Parameters::determineFileTypeFromFile(){
  ENTER("determineFileTypeFromFile");
  inputFile = fopen(inputFileName, "r");
  if(!inputFile){
    cerr << "File not found";
    exit(-1);
  }
  theFileType=assignmentsAndClassInformation;
  VERBOSE("File Type is ","Assignments And Class Information");
  EXIT("determineFileTypeFromFile");
  return theFileType;
}

int readline(FILE *fp, char *buf){
  ENTER("readline");
  int i = 0;
  while(!feof(fp)) {
    char c = fgetc(fp);
    if (c == '\n') {
      buf[i] = '\0';
      EXIT("readline");
      return i;
    } else {
      buf[i++] = c;
    }
  }
  buf[i] = '\0';
  EXIT("readline");
  return -1;
}


bool Parameters::verifyAndLoadFileStructure(int & locals,int & objects, int & fields){
  ENTER("verifyAndLoadFileStructure");
  determineFileTypeFromFile();
  char line[256];
  if(readline(inputFile,line)==-1){
    cerr << "Error reading locals: File format is \n Locals \n Objects \n Fields \n Assignment Statements ..." ;
    exit(-1);
  }
  locals=atoi(line);
  VERBOSE("Read Locals",locals);
  if(readline(inputFile,line)==-1){
    cerr << "Error reading objects: File format is \n Locals \n Objects \n Fields \n Assignment Statements ..." ;
    exit(-1);
  }
  objects=atoi(line);
  VERBOSE("Read Objects",objects);
  if(readline(inputFile,line)==-1){
    cerr << "Error reading Fields: File format is \n Locals \n Objects \n Fields \n Assignment Statements ..." ;
    exit(-1);
  }
  fields=atoi(line);	
  VERBOSE("Read Fields", fields);
  EXIT("verifyAndLoadFileStructure");
  return true;
}

void Parameters::prepareLoadInput() {
  currentInputFile = inputFile;
}
void Parameters::prepareLoadSolution(){
  solutionFile = fopen(solutionFileName, "r");
  currentInputFile = solutionFile;
}

// getValues function reads a line of input and parses the import according to current encoding
// if using default encoding, ENUMERATION, where each variable is enumerated
//    only p_v, q_v, f are used
// if using CLASS/METHOD/VARIABLE encoding, pc, pm, pv, qc, qm, qv, f are used 
AssignmentType Parameters::getValues(int &pc, int &pm, int &pv, int &qc, int &qm, int &qv, int &f){
  static AssignmentType type=unrecognized;
  ENTER("getValues");
  char line[256];

  if (currentInputFile == NULL)
    return nonAssignment;

  if(readline(currentInputFile,line)==-1)
    return nonAssignment;

  AssignmentType temp=getType(line);
  if (temp == Comments) {
    return Comments;
  } else if (temp!=Assignment){
    type=temp;
    if(readline(currentInputFile,line)==-1){
      return unrecognized;
    }
  }
  switch(type){
  case objectAllocation:
    //cout << 1;  
    if (varencoding == ENUMERATION) {
      //    obj, var
      sscanf(line, "%i %i",&pv, &qv);
    } else {       
      //    obj, cls, meth, var
      sscanf(line, "%i %i %i %i", &pv, &qc, &qm, &qv);
    }
    break;
  case directAssignment:
    //cout << 2;
    if (varencoding == ENUMERATION) {
      //    rhs, lhs
      sscanf(line, "%i %i", &pv, &qv);	
    } else {
      //    rcls, rmeth, rvar, lcls, lmeth, lvar
      sscanf(line, "%i %i %i %i %i %i", &pc, &pm, &pv, &qc, &qm, &qv);
    }
    break;
  case loadField:
    //cout << 3;
    if (varencoding == ENUMERATION) {
      //    rhs.f  lhs
      sscanf(line, "%i %i %i", &pv, &f, &qv);
    } else {
      sscanf(line, "%i %i %i %i %i %i %i", &pc, &pm, &pv, &f, &qc, &qm, &qv);
    }
    break;
  case fieldStore:
    //cout << 4;
    if (varencoding == ENUMERATION) {
      //   rhs, lhs.f
      sscanf(line, "%i %i %i", &pv, &qv, &f);
    } else {
      sscanf(line, "%i %i %i %i %i %i %i", &pc, &pm, &pv, &qc, &qm, &qv, &f);
    }
    break;
  case Solution:
    if (varencoding == ENUMERATION) {
      sscanf(line, "%i %i", &pv, &qv);
    } else {
      sscanf(line, "%i %i %i %i", &pc, &pm, &pv, &qv);
    }
    break;
  case ClassType:
    //cout << 5;
    sscanf(line, "%i %i", &pv, &qv);
    break;
  case AllocType:
    sscanf(line, "%i %i", &pv, &qv);
    //cout << 6;
    break;
  case VarType:
    //cout << 7;
    if (varencoding == ENUMERATION) {
      sscanf(line, "%i %i", &pv, &qv);
    } else {
      sscanf(line, "%i %i %i %i", &pc, &pm, &pv, &qv);
    }
    break;
  case FieldType:
    //cout << 8;
    // currently it is not used, format unknown.
    sscanf(line,"%i %i",&pv, &qv);
    break;
  case nonAssignment:
    //cout << 9;
    break;
  default:
    cerr << "Unrecognised Sequence found while determining Assignment Type";
    exit(-1);
  }
  EXIT("getValues");
  return type;
}

AssignmentType Parameters::getType(char * line){
  ENTER("getType");

  if(!strcmp(line,"Allocations:"))
    return objectAllocation;

  if(!strcmp(line,"Assignments:"))
    return directAssignment;
      
  if(!strcmp(line,"Stores:"))
    return fieldStore;

  if(!strcmp(line,"Loads:"))
    return loadField;

  if(!strcmp(line,"Declared Types:"))
    return ClassType;

  if(!strcmp(line,"Allocation Types:"))
    return AllocType;

  if(!strcmp(line,"Variable Types:"))
    return VarType;
  
  if(!strcmp(line,"Field Types:"))
    return FieldType;

  if(!strcmp(line,"Solution:"))
    return Solution;

  if(!strncmp(line, "#", 1)) {
    return Comments;

  }
  EXIT("getType");
  return Assignment;
}
