/* Solver-noalias.cxx: BDD solver of Andersen's points-to Analysis for Java
 * This is a specific version using no-alias, incremental/non-incremental,
 * and various orderings.
 *
 * Authors: Feng Qian (fqian@sable.mcgill.ca) 
 *          Marc Berndl(berndl@sable.mcgill.ca)
 */

#include <stdlib.h>
#include <sys/times.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include "fdd.h"
#include "helpers.h"

// noalias version uses five domains
const int NUMDOMAINS=5;
// V1 V2 are domains for variables 
// H1 H2 are domains for heap objects
// FD is a domain for field signature
int V1, V2, FD, H1, H2;
// T1 and T2 are used to compute typeFilter
// T1 = V2, and T2 = V1
int T1, T2;  

enum DOMORDER {IdxV1, IdxV2, IdxFD, IdxH1, IdxH2};
char * domainNames[] = {"V1", "V2", "FD", "H1", "H2"};

// the size of domains, can be changed to reflect the size of inputs
int domainBits[] = {18, 18, 13, 14, 14};
// to be computed in sysInit function
int domainSpos[] = {0, 0, 0, 0, 0}; 
// define the order of variables in each domain       
int * localOrders[] = {NULL, NULL, NULL, NULL, NULL};

// domain pairs for bdd_replace
bddPair * V1ToV2, * V2ToV1;
bddPair * H1ToH2, * H2ToH1;

// relations 
bdd pointsTo;     // V1 x H1
bdd edgeSet;      // V1 x V2
bdd typeFilter;   // V1 x H1
bdd stores;       // V1 x (V2 x FD) 
bdd loads;        // (V1 x FD) x V2

// solution for verification
bdd solutions;    // V1 x H1

// cached temoprary relations
bdd storePt, fieldPt, loadAss, loadPt;

void sysInit(Parameters * parms);
void varInit(Parameters * parms);
void loadInput(Parameters * parms);
void loadSolution(Parameters * parms);
void solve_incremental();
void solve_nonincremental();
void makeVarOrdering(VarOrdering ordering, bool reverseLocal, int varnum, int * varorder);
void remapping(int *varorder, int varnum, int *maps);
void reportOrdering(char*, int);
void getVariableMap(int*, int*, int);
int log2ceiling(int);
void generateRandomSequence(int num, int * rands);
void reportVarOrderingOfDomain(char * dname, int var);
int fillInVarIndices(int start, int * varorder, int numdoms, int *doms);

/* Main function processes parameters in Parameters.cxx, 
 * initialize BuDDy package and relations,
 * loads constraints from input files (in Parameters.cxx),
 * and solves the constraints. Verify results if it has one.
 */
int main(int argc, char ** args){
  Parameters * parms= new Parameters(argc,args);
  sysInit(parms);
  varInit(parms);
  loadInput(parms);

  if (INCREMENTAL == parms->incremental) {
    solve_incremental();
  } else {
    solve_nonincremental();
  }

  cout << "pointsTo has " << bdd_nodecount(pointsTo) << " nodes\n";

  // verifying results with solutions
  loadSolution(parms);
  if (solutions != bdd_false()) {
    cout << "solution has " << bdd_nodecount(solutions) << " nodes\n";
    if (solutions != pointsTo) {
      cout << "WRONG: result does not match the solution.\n";
    }
  }
}

/* initialize the BuDDy package */
void sysInit(Parameters * parms){

  int varnum, * varorder;

  // initialize bdd package first
  bdd_init(parms->numNodes, parms->cacheSize);
  bdd_setcacheratio(4);
  bdd_setmaxincrease(parms->cacheSize);
  
  int domains[NUMDOMAINS];
  for (int i=0; i<NUMDOMAINS; i++) {
    domains[i] = (1 << domainBits[i]);
  }
  fdd_extdomain(domains, NUMDOMAINS);
  V1 = IdxV1;
  V2 = IdxV2;
  FD = IdxFD;
  H1 = IdxH1;
  H2 = IdxH2;
  T1 = V2;
  T2 = V1;
  
  // get number of variables
  varnum= bdd_extvarnum(0);
  varorder = new int[varnum];

  makeVarOrdering((VarOrdering)parms->varordering, 
				  parms->reversedLocalOrdering, 
				  varnum, 
				  varorder);  

  bdd_setvarorder(varorder);
  bdd_enable_reorder();    // no reordering now

  reportOrdering(parms->orderingName, parms->varencoding);

  V1ToV2 = bdd_newpair(); V2ToV1 = bdd_newpair();
  H1ToH2 = bdd_newpair(); H2ToH1 = bdd_newpair();

  fdd_setpair(V1ToV2, V1, V2); fdd_setpair(V2ToV1, V2, V1);
  fdd_setpair(H1ToH2, H1, H2); fdd_setpair(H2ToH1, H2, H1);
}

/* report variable orderings */
void reportOrdering(char * ordername, int varencoding) {
  cout << ordername << " : \n";
 
  reportVarOrderingOfDomain("FD", FD);
  reportVarOrderingOfDomain("V1", V1);
  reportVarOrderingOfDomain("V2", V2);
  reportVarOrderingOfDomain("H1", H1);
  reportVarOrderingOfDomain("H2", H2); 
}

/* report variable orderings for single domain */
void reportVarOrderingOfDomain(char * dname, int var) {
  cout << dname << " : ";
  int vnum = fdd_varnum(var);
  int *vars = fdd_vars(var);
  for (int i=0; i<vnum; i++) {
    cout << bdd_var2level(vars[i]) << " ";
  }
  cout << "\n";
}

/* pull out variable map from BuDDy */
void getVariableMap(int * map, int *doms, int domnum) {
  int idx = 0;
  for (int var=0; var<domnum; var++) {
    int vnum = fdd_varnum(doms[var]);
    int *vars = fdd_vars(doms[var]);
    for (int i=0; i<vnum; i++) {
      map[idx++] = vars[i];
    }
  }
}

/* generate random sequences between [0 .. num) */
void generateRandomSequence(int num, int * rands) {
  srand(times(NULL));  // make a random seed
  for (int i=0; i<num; i++) {
    int r = rand()%num;
    // check if rand already used
    int j = 0;
    while (j<i) {
      if (rands[j] == r) {
	//used already
	r = rand()%num;
	j = 0;
      } else {
	j++;
      }
    }
    rands[i] = r;
  }
}

/* make variable orderings */
void makeVarOrdering(VarOrdering ordering, bool reverseLocal, 
					int varnum, int * varorder) {

  // compute domainBits from subdomainBits
  localOrders[IdxV1] = localOrders[IdxV2] = new int[domainBits[IdxV1]];
  localOrders[IdxH1] = localOrders[IdxH2] = new int[domainBits[IdxH1]];
  localOrders[IdxFD] = new int[domainBits[IdxFD]];
  

  // the local order should be adjusted later
  for (int i=0, pos=0; i<NUMDOMAINS; i++) {
    domainSpos[i] = pos;
    pos += domainBits[i];

    // by default, every domain has reversed order (most significant bit first)
    for (int j=0; j<domainBits[i]; j++) {
      if (reverseLocal) {
        localOrders[i][j] = domainBits[i] - j - 1;
      } else {
	localOrders[i][j] = j;
      }
    }
  }

  int * doms = new int[NUMDOMAINS];

  switch (ordering) {
  case FD_V1_V2_H1_H2:
    {
      int idx = 0;
      doms[0] = IdxFD;
      idx = fillInVarIndices(idx, varorder, 1, doms);
      doms[0] = IdxV1;
      idx = fillInVarIndices(idx, varorder, 1, doms);
      doms[0] = IdxV2;
      idx = fillInVarIndices(idx, varorder, 1, doms);
      doms[0] = IdxH1;
      idx = fillInVarIndices(idx, varorder, 1, doms);
      doms[0] = IdxH2;
      idx = fillInVarIndices(idx, varorder, 1, doms);      
    }
    break;
  case FD_V1xV2_H1_H2:
    {
      int idx = 0;
      doms[0] = IdxFD;
      idx = fillInVarIndices(idx, varorder, 1, doms);
      doms[0] = IdxV1; doms[1] = IdxV2;
      idx = fillInVarIndices(idx, varorder, 2, doms);
      doms[0] = IdxH1; 
      idx = fillInVarIndices(idx, varorder, 1, doms);
      doms[0] = IdxH2;
      idx = fillInVarIndices(idx, varorder, 1, doms);
    }
    break;
  case FD_V2_V1_H1_H2:
    {
      int idx = 0;
      doms[0] = IdxFD;
      idx = fillInVarIndices(idx, varorder, 1, doms);
      doms[0] = IdxV2; 
      idx = fillInVarIndices(idx, varorder, 1, doms);
      doms[0] = IdxV1;
      idx = fillInVarIndices(idx, varorder, 1, doms);
      doms[0] = IdxH1; 
      idx = fillInVarIndices(idx, varorder, 1, doms);
      doms[0] = IdxH2;
      idx = fillInVarIndices(idx, varorder, 1, doms);
    }
    break;
  case FD_V2xV1_H1_H2:
    {
      int idx = 0;
      doms[0] = IdxFD;
      idx = fillInVarIndices(idx, varorder, 1, doms);
      doms[0] = IdxV2; doms[1] = IdxV1;
      idx = fillInVarIndices(idx, varorder, 2, doms);
      doms[0] = IdxH1; 
      idx = fillInVarIndices(idx, varorder, 1, doms);
      doms[0] = IdxH2;
      idx = fillInVarIndices(idx, varorder, 1, doms);      
    }
    break;
  case FDxV1xV2xH1xH2:
    {
      int idx = 0;
      doms[0] = IdxFD; doms[1] = IdxV1; doms[2] = IdxV2;
      doms[3] = IdxH1; doms[4] = IdxH2;
      idx = fillInVarIndices(idx, varorder, 5, doms);
    }
    break;
  case FD_H1_V1xV2_H2:
    {
      int idx = 0;
      doms[0] = IdxFD;
      idx = fillInVarIndices(idx, varorder, 1, doms);
      doms[0] = IdxH1; 
      idx = fillInVarIndices(idx, varorder, 1, doms);
      doms[0] = IdxV1; doms[1] = IdxV2;
      idx = fillInVarIndices(idx, varorder, 2, doms);
      doms[0] = IdxH2;
      idx = fillInVarIndices(idx, varorder, 1, doms);      
    }
    break;
  case FD_H1_V2xV1_H2:
    {
      int idx = 0;
      doms[0] = IdxFD;
      idx = fillInVarIndices(idx, varorder, 1, doms);
      doms[0] = IdxH1; 
      idx = fillInVarIndices(idx, varorder, 1, doms);
      doms[0] = IdxV2; doms[1] = IdxV1;
      idx = fillInVarIndices(idx, varorder, 2, doms);
      doms[0] = IdxH2;
      idx = fillInVarIndices(idx, varorder, 1, doms);      
    }
    break;
  case H1_V1xV2_FD_H2:
    {
      int idx = 0;
      doms[0] = IdxH1;
      idx = fillInVarIndices(idx, varorder, 1, doms);
      doms[0] = IdxV1; doms[1] = IdxV2;
      idx = fillInVarIndices(idx, varorder, 2, doms);
      doms[0] = IdxFD;
      idx = fillInVarIndices(idx, varorder, 1, doms);
      doms[0] = IdxH2;
      idx = fillInVarIndices(idx, varorder, 1, doms);
    }
    break;
  case H1_V2xV1_FD_H2:
    {
      int idx = 0;
      doms[0] = IdxH1;
      idx = fillInVarIndices(idx, varorder, 1, doms);
      doms[0] = IdxV2; doms[1] = IdxV1;
      idx = fillInVarIndices(idx, varorder, 2, doms);
      doms[0] = IdxFD;
      idx = fillInVarIndices(idx, varorder, 1, doms);
      doms[0] = IdxH2;
      idx = fillInVarIndices(idx, varorder, 1, doms);
    }
  case FD_V1xV2xH1_H2:
    {
      int idx = 0;
      doms[0] = IdxFD;
      idx = fillInVarIndices(idx, varorder, 1, doms);
      doms[0] = IdxV1; doms[1] = IdxV2; doms[2] = IdxH1;
      idx = fillInVarIndices(idx, varorder, 3, doms);
      doms[0] = IdxH2;
      idx = fillInVarIndices(idx, varorder, 1, doms);
    }
    break;
  default:
    cerr << "Ordering is not defined\n";
    exit(-1);
    break;
  }

  // according to the documentation of buddy, the default ordering is x1, y1, z1, x2, y2, z2, .....
  // V1[0] -> default variable number
  int *outside2inside = new int[varnum];
  doms[IdxV1] = V1; doms[IdxV2] = V2;
  doms[IdxFD] = FD; doms[IdxH1] = H1; doms[IdxH2] = H2;
  getVariableMap(outside2inside, doms, NUMDOMAINS);

  remapping(varorder, varnum, outside2inside);

  delete(outside2inside);
  delete(doms);
  delete(localOrders[IdxV1]); delete(localOrders[IdxFD]); delete(localOrders[IdxH1]);
}

/* remap according to a map */
void remapping(int *varorder, int varnum, int *maps) {
  for (int i=0; i<varnum; i++) {
    varorder[i] = maps[varorder[i]];
  }
}

int fillInVarIndices(int start, int * varorder, int numdoms, int *doms) {
  int totalvars = 0;
  int *bits = new int[numdoms];
  for (int i=0; i<numdoms; i++) {
    totalvars += domainBits[doms[i]];
    bits[i] = 0;
  }

  for (int i=start, n=start+totalvars, j=0; i<n; i++) {
    while (bits[j] >= domainBits[doms[j]]) {
      j = (j + 1) % numdoms;
    }
    varorder[i] = domainSpos[doms[j]] + localOrders[doms[j]][bits[j]++];
    j = (j + 1) % numdoms;
  }

  delete(bits);
  return start+totalvars;
}

void varInit(Parameters * parms) {
  pointsTo = bdd_false();
  edgeSet = bdd_false();
  typeFilter = bdd_false();
  stores = bdd_false();
  loads = bdd_false();
  solutions = bdd_false();
  storePt = bdd_false(); loadPt = bdd_false(); loadAss = bdd_false(); fieldPt = bdd_false();
}

void loadInput(Parameters * parms){
  // profiling
  int maxcls = 0, maxmid = 0, maxvar = 0, maxfld = 0, maxalc = 0;
  
  bool eof = false;
  bdd cC = bdd_false(), aC = bdd_false(), vC = bdd_false(), fC = bdd_false();

  parms->determineFileTypeFromFile();  
  parms->prepareLoadInput();

  clock_t ct1= clock()/10000;
  
  while(!eof){
    int pc=0, pm=0, pv=0, qc=0, qm=0, qv=0, f=0;
    AssignmentType atype = parms->getValues(pc, pm, pv, qc, qm, qv, f);

    switch(atype){
    case objectAllocation:  // q = alloc_p
      pointsTo |= (fdd_ithvar(H1, -pv) & fdd_ithvar(V1, qv));

      if ((-pv) > maxalc) maxalc = -pv;
      if (qm > maxmid) maxmid = qm;
      if (pv > maxvar) maxvar = pv;
      if (qv > maxvar) maxvar = qv;
      
      break;
    case directAssignment:     // q = p
      edgeSet |= (fdd_ithvar(V1, pv) & fdd_ithvar(V2, qv));

      if (pc > maxcls) maxcls = pc;
      if (qc > maxcls) maxcls = qc;
      if (pm > maxmid) maxmid = pm;
      if (qm > maxmid) maxmid = qm;
      if (pv > maxvar) maxvar = pv;
      if (qv > maxvar) maxvar = qv;
     
      break;
    case loadField:   // q = p.f   V2 = V1 x F
      loads |= (fdd_ithvar(V2, qv) & fdd_ithvar(V1, pv) & fdd_ithvar(FD, f));

      if (f > maxfld) maxfld = f;
      if (pc > maxcls) maxcls = pc;
      if (qc > maxcls) maxcls = qc;
      if (pm > maxmid) maxmid = pm;
      if (qm > maxmid) maxmid = qm;
      if (pv > maxvar) maxvar = pv;
      if (qv > maxvar) maxvar = qv;
      
      break;
    case fieldStore:  // q.f = p   V2 x F = V1
      stores |= (fdd_ithvar(V2, qv) & fdd_ithvar(FD, f) & fdd_ithvar(V1, pv));
	
      if (f > maxfld) maxfld = f;      
      if (pc > maxcls) maxcls = pc;
      if (qc > maxcls) maxcls = qc;
      if (pm > maxmid) maxmid = pm;
      if (qm > maxmid) maxmid = qm;
      if (pv > maxvar) maxvar = pv;
      if (qv > maxvar) maxvar = qv;
      
      break;
    case ClassType:   // T1 x T2
      cC |= (fdd_ithvar(T1, pv) & fdd_ithvar(T2, qv));
      break;
    case AllocType:   // H x T2
      aC |= (fdd_ithvar(H1, -pv) & fdd_ithvar(T2, qv));
      break;
    case VarType:     // V1 x T1
      vC |= (fdd_ithvar(V1, pv) & fdd_ithvar(T1, qv));
      break;
    case Solution:    // p = -q
      solutions |= (fdd_ithvar(H1, -qv) & fdd_ithvar(V1, pv));
      break;
    case Comments:
      break;
    default:
      // reach the end of file 
      eof = true;
      break;
    }
  }  
  clock_t ct2 = clock()/10000;
  cout << "read input : " << (ct2-ct1)*0.01 << "s\n";  

  // prepare a few constants
  clock_t tt1 = clock()/10000;
  if (!(parms->typeFilter)) {
    typeFilter = bdd_true();
  } else {
    // (T1 x T2) * (H1 x T2) => (T1 x H1)
    bdd assignableTypes = bdd_relprod(cC, aC, fdd_ithset(T2));
    // (T1 x H1) * (V1 x T1) => (V1 x H1)
    typeFilter = bdd_relprod(assignableTypes, vC, fdd_ithset(T1));
  }
  clock_t tt2 = clock()/10000;
  cout << "build type filter : " << (tt2 - tt1)*0.01 << "s\n";
  cout << "type filter size  : " << bdd_nodecount(typeFilter) << " nodes\n";
  
  cout << "loads has  " << bdd_nodecount(loads) << " nodes\n";
  cout << "stores has " << bdd_nodecount(stores) << " nodes\n";

  clock_t ct3 = clock()/10000;
  cout << "LOADING TIME: " << (ct3-ct1)*0.01 << "s\n";  
  cout << "edgeSet has " << bdd_nodecount(edgeSet) << " nodes\n";
  cout << "pointsTo has " << bdd_nodecount(pointsTo) << " nodes\n";

  cout << "\n";
  cout << "max vars      " << maxvar << "(" << log2ceiling(maxvar+1) << ")\n";
  cout << "max fields    " << maxfld << "(" << log2ceiling(maxfld+1) << ")\n";
  cout << "max allocs    " << maxalc << "(" << log2ceiling(maxalc+1) << ")\n";

  // verify the input is type-safe
  if (pointsTo != (pointsTo & typeFilter)) {
    cout << "pointsTo & typeFilter results " << bdd_nodecount(pointsTo & typeFilter) << " nodes\n";
    exit (-1);
  }
}

void loadSolution(Parameters * parms) {
  parms->prepareLoadSolution();

  bool eof = false;
  int pc, pm, pv, qc, qm, qv, f;
  while (!eof) {
    AssignmentType atype = parms->getValues(pc, pm, pv, qc, qm, qv, f);
    //    assert(parms->varencoding == CLSMETHVAR);
    switch (atype) {
    case Solution:
      solutions |= (fdd_ithvar(H1, -qv) & fdd_ithvar(V1, pv));
      break;
    case Comments:
      break;
    default:
      eof = true;
      break;
    }
  }
}


void solve_nonincremental() {
  clock_t innerLoopPropTime=0, outloopPropTime=0;
  clock_t ct1, ct2, tt1, tt2;
  bdd oldPt1, oldPt2;

  tt1 = clock()/10000;
  // start solving 
  do{
    oldPt1 = pointsTo;
    ct1 = clock()/10000;
    // repeat rule (1) in the inner loop
    bdd oldPt2 = bdd_false();
    do{
      oldPt2 = pointsTo;
      /* --- rule (1) --- */
      // 
      //   l1 -> l2    o \in pt(l1)
      // --------------------------
      //          o \in pt(l2)
      bdd newPt1 = bdd_relprod(edgeSet, pointsTo, fdd_ithset(V1));
      bdd newPt2 = bdd_replace(newPt1, V2ToV1);

      /* --- apply type filtering and merge into pointsTo relation --- */
      bdd newPt3 = newPt2 & typeFilter;
      pointsTo = pointsTo | newPt3;
	
    }while(oldPt2 != pointsTo);

    ct2 = clock()/10000;
    innerLoopPropTime +=(ct2-ct1);

    // propagate points-to set over field loads and stores
    ct1 = clock()/10000;
    /* --- rule (2) --- */
    //
    //   o2 \in pt(l)   l -> q.f   o1 \in pt(q)
    // -----------------------------------------
    //                  o2 \in pt(o1.f) 
    bdd tmpRel1 = bdd_relprod(stores, pointsTo, fdd_ithset(V1));     // (V2xFD)xH1
    bdd tmpRel2 = bdd_replace(bdd_replace(tmpRel1, V2ToV1), H1ToH2); // (V1xFD)xH2
        fieldPt = bdd_relprod(tmpRel2, pointsTo, fdd_ithset(V1));    // (H1xFD)xH2
      
    /* --- rule (3) --- */
    //
    //   p.f -> l   o1 \in pt(p)   o2 \in pt(o1)
    // -----------------------------------------
    //                 o2 \in pt(l)
    bdd tmpRel3 = bdd_relprod(loads, pointsTo, fdd_ithset(V1));       // (H1xFD)xV2
    bdd newPt4  = bdd_relprod(tmpRel3, fieldPt, fdd_ithset(H1)&fdd_ithset(FD)); // V2xH2
    bdd newPt5  = bdd_replace(bdd_replace(newPt4, V2ToV1), H2ToH1);   // V1xH2
     
    /* --- apply type filtering and merge into pointsTo relation --- */
    bdd newPt6  = newPt5 & typeFilter;
    pointsTo = pointsTo | newPt6;

    ct2 = clock()/10000;
    outloopPropTime += (ct2 - ct1);

  }while(oldPt1 != pointsTo);

  tt2 = clock()/10000;
  
  cout << "\n";
  cout << "innerloop prop time  " << innerLoopPropTime* 0.01 << "s\n";
  cout << "outloop prop time    " << outloopPropTime*0.01 << "s\n";
  cout << "SOLVING TIME: " << (tt2-tt1)*0.01 << "s\n";
}

/* incrementally computes points-to relation
 *    oldPt = A x B
 *    newPt = (A + A') x B 
 *          = A x B + A' x B
 *          = oldPt + A' x B
 *    therefore, we only needs to compute A' x B 
 *    note : x is relprod here, expensive operation
 */
void solve_incremental(){
  clock_t innerLoopPropTime=0, outloopPropTime=0;
  clock_t ct1, ct2, tt1, tt2;

  bdd oldPointsTo = bdd_false();
  bdd newPointsTo = pointsTo;

  // start solving 
  tt1 = clock()/10000;
  do{
    ct1 = clock()/10000;
    
    // repeat rule (1) in the inner loop
    do {
      bdd newPt1 = bdd_relprod(edgeSet, newPointsTo, fdd_ithset(V1));
      bdd newPt2 = bdd_replace(newPt1, V2ToV1);
      bdd newPt3 = newPt2 - pointsTo;
      newPointsTo = newPt3 & typeFilter;
      pointsTo = pointsTo | newPointsTo;
    } while (newPointsTo != bdd_false());

    ct2 = clock()/10000;
    innerLoopPropTime +=(ct2-ct1);

    newPointsTo = pointsTo - oldPointsTo;

    ct1 = clock()/10000;
    // apply rule (2)
    bdd tmpRel1 = bdd_relprod(stores, newPointsTo, fdd_ithset(V1));     // (V2xFD)xH1
    bdd tmpRel2 = bdd_replace(bdd_replace(tmpRel1, V2ToV1), H1ToH2);    // (V1xFD)xH2
    bdd newStorePt = tmpRel2 - storePt; 	
    // cache storePt
    storePt |= newStorePt;                                              // (V1xFD)xH2
      
    bdd newFieldPt = bdd_relprod(storePt, newPointsTo, fdd_ithset(V1)); // (H1xFD)xH2 
    newFieldPt |= bdd_relprod(newStorePt, oldPointsTo, fdd_ithset(V1)); // (H1xFD)xH2
    newFieldPt -= fieldPt; 
    // cache fieldPt
    fieldPt |= newFieldPt;                                              // (H1xFD)xH2
      
    // apply rule (3)
    bdd tmpRel3 = bdd_relprod(loads, newPointsTo, fdd_ithset(V1));      // (H1xFD)xV2
    bdd newLoadAss = tmpRel3 - loadAss; 
    bdd newLoadPt  = bdd_relprod(loadAss, newFieldPt, fdd_ithset(H1)&fdd_ithset(FD)); // V2xH2 
    newLoadPt |= bdd_relprod(newLoadAss, fieldPt, fdd_ithset(H1)&fdd_ithset(FD));     // V2xH2
    // cache loadAss
    loadAss |= newLoadAss; 
      
    // update oldPointsTo
    oldPointsTo = pointsTo; 

    // convert new points-to relation to normal type
    newPointsTo = bdd_replace(bdd_replace(newLoadPt, V2ToV1), H2ToH1);
    newPointsTo -= pointsTo;
    
    // apply typeFilter
    newPointsTo = typeFilter & newPointsTo;
    pointsTo |= newPointsTo;

    ct2 = clock()/10000;
    outloopPropTime += (ct2 - ct1);

  }while(newPointsTo != bdd_false());

  tt2 = clock()/10000;

  cout << "\n";
  cout << "innerloop prop time  " << innerLoopPropTime* 0.01 << "s\n";
  cout << "outloop prop time    " << outloopPropTime*0.01 << "s\n";
  cout << "SOLVING TIME: " << (tt2-tt1)*0.01 << "s\n";
}

// return how many useful bits in v
int log2ceiling(int v) {
  int bits = 0;
  while (v != 0) {
    bits++;
    v >>= 1;
  }
  return bits;
}

